/** 
 * DevExtreme (ui/drop_down_editor/ui.drop_down_list.js)
 * Version: 16.1.17
 * Build date: Tue Mar 12 2019
 *
 * Copyright (c) 2012 - 2019 Developer Express Inc. ALL RIGHTS RESERVED
 * EULA: https://www.devexpress.com/Support/EULAs/DevExtreme.xml
 * MAY BE USED WITH: DevExtreme Complete, DevExtreme Mobile, DevExtreme Web
 */
"use strict";
define(function(require, exports, module) {
    var $ = require("jquery"),
        Guid = require("../../core/guid"),
        registerComponent = require("../../core/component_registrator"),
        commonUtils = require("../../core/utils/common"),
        DropDownEditor = require("./ui.drop_down_editor"),
        List = require("../list"),
        errors = require("../widget/ui.errors"),
        eventUtils = require("../../events/utils"),
        devices = require("../../core/devices"),
        DataExpressionMixin = require("../editor/ui.data_expression"),
        messageLocalization = require("../../localization/message"),
        themes = require("../themes");
    var LIST_ITEM_SELECTOR = ".dx-list-item",
        LIST_ITEM_DATA_KEY = "dxListItemData",
        DROPDOWNLIST_POPUP_WRAPPER_CLASS = "dx-dropdownlist-popup-wrapper",
        SKIP_GESTURE_EVENT_CLASS = "dx-skip-gesture-event",
        SEARCH_MODES = ["startswith", "contains", "endwith", "notcontains"];
    var DropDownList = DropDownEditor.inherit({
        _supportedKeys: function() {
            var parent = this.callBase();
            return $.extend({}, parent, {
                tab: function(e) {
                    if (this.option("opened") && "instantly" === this.option("applyValueMode")) {
                        var $focusedItem = this._list.option("focusedElement");
                        $focusedItem && this._setSelectedElement($focusedItem)
                    } else {
                        this._focusTarget().focusout()
                    }
                    parent.tab.apply(this, arguments)
                },
                space: $.noop,
                home: $.noop,
                end: $.noop
            })
        },
        _setSelectedElement: function($element) {
            var value = this._valueGetter(this._list._getItemData($element));
            this._setValue(value)
        },
        _setValue: function(value) {
            this.option("value", value)
        },
        _setDeprecatedOptions: function() {
            this.callBase();
            $.extend(this._deprecatedOptions, {
                pagingEnabled: {
                    since: "15.1",
                    message: "Use the 'dataSource.paginate' option instead"
                }
            })
        },
        _getDefaultOptions: function() {
            return $.extend(this.callBase(), $.extend(DataExpressionMixin._dataExpressionDefaultOptions(), {
                displayValue: void 0,
                searchEnabled: false,
                searchMode: "contains",
                searchTimeout: 500,
                minSearchLength: 0,
                searchExpr: null,
                valueChangeEvent: "input change keyup",
                selectedItem: null,
                pagingEnabled: void 0,
                noDataText: messageLocalization.format("dxCollectionWidget-noDataText"),
                onSelectionChanged: null,
                onItemClick: $.noop,
                showDataBeforeSearch: false,
                popupPosition: {
                    my: "left top",
                    at: "left bottom",
                    offset: {
                        h: 0,
                        v: 0
                    },
                    collision: "flip"
                },
                popupWidthExtension: 0
            }))
        },
        _defaultOptionsRules: function() {
            return this.callBase().concat([{
                device: function(device) {
                    return "win" === device.platform && device.version && 8 === device.version[0]
                },
                options: {
                    popupPosition: {
                        offset: {
                            v: -6
                        }
                    }
                }
            }, {
                device: function() {
                    return /android5/.test(themes.current())
                },
                options: {
                    popupWidthExtension: 32
                }
            }, {
                device: {
                    platform: "ios"
                },
                options: {
                    popupPosition: {
                        offset: {
                            v: -1
                        }
                    }
                }
            }, {
                device: {
                    platform: "generic"
                },
                options: {
                    buttonsLocation: "bottom center"
                }
            }])
        },
        _setOptionsByReference: function() {
            this.callBase();
            $.extend(this._optionsByReference, {
                value: true,
                selectedItem: true,
                displayValue: true
            })
        },
        _init: function() {
            this.callBase();
            this._initDataExpressions();
            this._initActions();
            this._setListDataSource();
            this._validateSearchMode();
            this._clearSelectedItem()
        },
        _initActions: function() {
            this._initContentReadyAction();
            this._initSelectionChangedAction();
            this._initItemClickAction()
        },
        _initContentReadyAction: function() {
            this._contentReadyAction = this._createActionByOption("onContentReady", {
                excludeValidators: ["disabled", "readOnly"]
            })
        },
        _initSelectionChangedAction: function() {
            this._selectionChangedAction = this._createActionByOption("onSelectionChanged", {
                excludeValidators: ["disabled", "readOnly"]
            })
        },
        _initItemClickAction: function() {
            this._itemClickAction = this._createActionByOption("onItemClick")
        },
        _renderField: function() {
            this.callBase();
            this._input().on("input", $.proxy(this._setFocusPolicy, this))
        },
        _preventFocusOnPopup: function(e) {
            if (this._list && this._list.initialOption("focusStateEnabled")) {
                e.preventDefault()
            }
        },
        _createPopup: function() {
            this.callBase();
            this._popup._wrapper().addClass(this._popupWrapperClass());
            this._popup.content().off("mousedown").on("mousedown", $.proxy(this._preventFocusOnPopup, this))
        },
        _popupWrapperClass: function() {
            return DROPDOWNLIST_POPUP_WRAPPER_CLASS
        },
        _renderInputValue: function() {
            var callBase = $.proxy(this.callBase, this),
                value = this._getCurrentValue();
            return this._loadItem(value).always($.proxy(function(item) {
                this._setSelectedItem(item);
                callBase(value)
            }, this))
        },
        _loadItem: function(value) {
            var selectedItem = $.grep(this.option("items") || [], $.proxy(function(item) {
                return this._isValueEquals(this._valueGetter(item), value)
            }, this))[0];
            return void 0 !== selectedItem ? $.Deferred().resolve(selectedItem).promise() : this._loadValue(value)
        },
        _setSelectedItem: function(item) {
            var displayValue = this._displayValue(item);
            this.option("selectedItem", commonUtils.ensureDefined(item, null));
            this.option("displayValue", displayValue)
        },
        _displayValue: function(item) {
            return this._displayGetter(item)
        },
        _refreshSelected: function() {
            this._listItemElements().each($.proxy(function(_, itemElement) {
                var $itemElement = $(itemElement);
                var itemValue = this._valueGetter($itemElement.data(LIST_ITEM_DATA_KEY));
                var isItemSelected = this._isSelectedValue(itemValue);
                if (isItemSelected) {
                    this._list.selectItem($itemElement)
                } else {
                    this._list.unselectItem($itemElement)
                }
            }, this))
        },
        _popupShownHandler: function() {
            this.callBase();
            this._setFocusPolicy()
        },
        _setFocusPolicy: function() {
            if (!this.option("focusStateEnabled") || !this._list) {
                return
            }
            this._list.option("focusedElement", null)
        },
        _isSelectedValue: function(value) {
            return this._isValueEquals(value, this.option("value"))
        },
        _validateSearchMode: function() {
            var searchMode = this.option("searchMode"),
                normalizedSearchMode = searchMode.toLowerCase();
            if ($.inArray(normalizedSearchMode, SEARCH_MODES) < 0) {
                throw errors.Error("E1019", searchMode)
            }
        },
        _clearSelectedItem: function() {
            this.option("selectedItem", null)
        },
        _processDataSourceChanging: function() {
            this._setListDataSource();
            this._renderInputValue().fail($.proxy(function() {
                if (this._isCustomValueAllowed()) {
                    return
                }
                this._clearSelectedItem()
            }, this))
        },
        _isCustomValueAllowed: function() {
            return this.option("displayCustomValue")
        },
        reset: function() {
            this.callBase();
            this._clearFilter();
            this._clearSelectedItem()
        },
        _listItemElements: function() {
            return this._$list ? this._$list.find(LIST_ITEM_SELECTOR) : $()
        },
        _popupConfig: function() {
            var that = this;
            return $.extend(this.callBase(), {
                templatesRenderAsynchronously: false,
                width: this.option("width"),
                onShowing: function() {
                    that.element().addClass(SKIP_GESTURE_EVENT_CLASS)
                },
                onHidden: function() {
                    that.element().removeClass(SKIP_GESTURE_EVENT_CLASS)
                },
                height: "auto",
                maxHeight: $.proxy(this._getMaxHeight, this)
            })
        },
        _renderPopupContent: function() {
            this._renderList()
        },
        _attachChildKeyboardEvents: function() {
            this._childKeyboardProcessor = this._keyboardProcessor.attachChildProcessor();
            this._setListOption("_keyboardProcessor", this._childKeyboardProcessor)
        },
        _fireContentReadyAction: $.noop,
        _setAriaTargetForList: function() {
            this._list._getAriaTarget = $.proxy(this._getAriaTarget, this);
            this._list.setAria("role", "combobox")
        },
        _renderList: function() {
            this._listId = (new Guid)._value;
            var $list = this._$list = $("<div>", {
                id: this._listId
            }).appendTo(this._popup.content());
            this._list = this._createComponent($list, List, this._listConfig());
            this._refreshList();
            this._setAriaTargetForList()
        },
        _renderOpenedState: function() {
            this.callBase();
            var opened = this.option("opened") || void 0;
            this.setAria({
                activedescendant: opened && this._list.getFocusedItemId(),
                owns: opened && this._listId
            })
        },
        _refreshList: function() {
            if (this._list && this._shouldRefreshDataSource()) {
                this._setListDataSource()
            }
        },
        _shouldRefreshDataSource: function() {
            var dataSourceProvided = !!this._list.option("dataSource");
            return dataSourceProvided !== this._needPassDataSourceToList()
        },
        _isDesktopDevice: function() {
            return "desktop" === devices.real().deviceType
        },
        _listConfig: function() {
            return {
                selectionMode: "single",
                _templates: this.option("_templates"),
                templateProvider: this.option("templateProvider"),
                noDataText: this.option("noDataText"),
                onContentReady: $.proxy(this._listContentReadyHandler, this),
                itemTemplate: this._getTemplateByOption("itemTemplate"),
                indicateLoading: false,
                tabIndex: -1,
                onItemClick: $.proxy(this._listItemClickAction, this),
                dataSource: this._getDataSource(),
                _keyboardProcessor: this._childKeyboardProcessor,
                hoverStateEnabled: this._isDesktopDevice() ? this.option("hoverStateEnabled") : false,
                focusStateEnabled: this._isDesktopDevice() ? this.option("focusStateEnabled") : false,
                selectedItem: this.option("selectedItem")
            }
        },
        _getDataSource: function() {
            return this._needPassDataSourceToList() ? this._dataSource : null
        },
        _dataSourceOptions: function() {
            this._suppressDeprecatedWarnings();
            var pagingEnabled = this.option("pagingEnabled");
            this._resumeDeprecatedWarnings();
            return {
                paginate: commonUtils.ensureDefined(pagingEnabled, false)
            }
        },
        _listContentReadyHandler: function() {
            this._list = this._list || this._$list.dxList("instance");
            this.option().items = this._list.option("items");
            this._dimensionChanged();
            this._contentReadyAction()
        },
        _setListOption: function(optionName, value) {
            this._setWidgetOption("_list", arguments)
        },
        _listItemClickAction: function(e) {
            this._listItemClickHandler(e);
            this._itemClickAction(e)
        },
        _listItemClickHandler: $.noop,
        _setListDataSource: function() {
            if (!this._list) {
                return
            }
            this._setListOption("dataSource", this._getDataSource());
            if (!this._needPassDataSourceToList()) {
                this._setListOption("items", [])
            }
        },
        _needPassDataSourceToList: function() {
            return this.option("showDataBeforeSearch") || this._isMinSearchLengthExceeded()
        },
        _isMinSearchLengthExceeded: function() {
            return this._searchValue().toString().length >= this.option("minSearchLength")
        },
        _searchValue: function() {
            return this._input().val() || ""
        },
        _getSearchEvent: function() {
            return eventUtils.addNamespace("keyup", this.NAME + "Search")
        },
        _renderEvents: function() {
            this.callBase();
            if (this._shouldRenderSearchEvent()) {
                this._input().on(this._getSearchEvent(), $.proxy(this._searchHandler, this))
            }
        },
        _shouldRenderSearchEvent: function() {
            return this.option("searchEnabled")
        },
        _refreshEvents: function() {
            this._input().off(this._getSearchEvent());
            this.callBase()
        },
        _searchHandler: function(e) {
            if (!this._isMinSearchLengthExceeded()) {
                this._searchCanceled();
                return
            }
            var searchTimeout = this.option("searchTimeout");
            if (searchTimeout) {
                this._clearSearchTimer();
                this._searchTimer = setTimeout($.proxy(this._searchDataSource, this), searchTimeout)
            } else {
                this._searchDataSource()
            }
        },
        _searchCanceled: function() {
            this._clearSearchTimer();
            if (this._needPassDataSourceToList()) {
                this._filterDataSource(null)
            }
            this._refreshList()
        },
        _searchDataSource: function() {
            this._filterDataSource(this._searchValue())
        },
        _filterDataSource: function(searchValue) {
            this._clearSearchTimer();
            var dataSource = this._dataSource;
            dataSource.searchExpr(this.option("searchExpr") || this._displayGetterExpr());
            dataSource.searchOperation(this.option("searchMode"));
            dataSource.searchValue(searchValue);
            return dataSource.load().done($.proxy(this._dataSourceFiltered, this, searchValue))
        },
        _clearFilter: function() {
            var dataSource = this._dataSource;
            dataSource && dataSource.searchValue() && dataSource.searchValue(null)
        },
        _dataSourceFiltered: function() {
            this._refreshList();
            this._refreshPopupVisibility()
        },
        _refreshPopupVisibility: function() {
            if (this.option("readOnly")) {
                return
            }
            this.option("opened", this._hasItemsToShow());
            if (this.option("opened")) {
                this._dimensionChanged()
            }
        },
        _hasItemsToShow: function() {
            var resultItems = this._dataSource && this._dataSource.items() || [];
            var resultAmount = resultItems.length;
            var isMinSearchLengthExceeded = this._needPassDataSourceToList();
            return isMinSearchLengthExceeded && resultAmount && this._hasFocusClass()
        },
        _clearSearchTimer: function() {
            clearTimeout(this._searchTimer);
            delete this._searchTimer
        },
        _popupShowingHandler: function() {
            this._dimensionChanged()
        },
        _dimensionChanged: function() {
            this._popup && this._updatePopupDimensions()
        },
        _updatePopupDimensions: function() {
            this._updatePopupWidth();
            this._updatePopupHeight()
        },
        _updatePopupWidth: function() {
            this._setPopupOption("width", this.element().outerWidth() + this.option("popupWidthExtension"))
        },
        _needPopupRepaint: function() {
            var currentPageIndex = this._dataSource.pageIndex(),
                needRepaint = commonUtils.isDefined(this._pageIndex) && currentPageIndex <= this._pageIndex;
            this._pageIndex = currentPageIndex;
            return needRepaint
        },
        _updatePopupHeight: function() {
            if (this._needPopupRepaint()) {
                this._popup.repaint()
            }
            this._list && this._list.updateDimensions()
        },
        _getMaxHeight: function() {
            var $element = this.element(),
                offset = $element.offset(),
                windowHeight = $(window).height(),
                maxHeight = Math.max(offset.top, windowHeight - offset.top - $element.outerHeight());
            return Math.min(.5 * windowHeight, maxHeight)
        },
        _clean: function() {
            if (this._list) {
                delete this._list
            }
            this.callBase()
        },
        _dispose: function() {
            this._clearSearchTimer();
            this.callBase()
        },
        _setCollectionWidgetOption: function() {
            this._setListOption.apply(this, arguments)
        },
        _optionChanged: function(args) {
            this._dataExpressionOptionChanged(args);
            switch (args.name) {
                case "hoverStateEnabled":
                case "focusStateEnabled":
                    this._isDesktopDevice() && this._setListOption(args.name, args.value);
                    this.callBase(args);
                    break;
                case "items":
                    if (!this.option("dataSource")) {
                        this._processDataSourceChanging()
                    }
                    break;
                case "dataSource":
                    this._processDataSourceChanging();
                    break;
                case "valueExpr":
                case "displayExpr":
                    this._renderValue();
                    break;
                case "searchMode":
                    this._validateSearchMode();
                    break;
                case "minSearchLength":
                    this._refreshList();
                    break;
                case "searchEnabled":
                case "showDataBeforeSearch":
                case "searchExpr":
                case "pagingEnabled":
                    this._invalidate();
                    break;
                case "onContentReady":
                    this._initContentReadyAction();
                    break;
                case "onSelectionChanged":
                    this._initSelectionChangedAction();
                    break;
                case "onItemClick":
                    this._initItemClickAction();
                    break;
                case "noDataText":
                    this._setListOption("noDataText");
                    break;
                case "displayValue":
                    this.option("text", args.value);
                    break;
                case "itemTemplate":
                case "searchTimeout":
                case "popupWidthExtension":
                    break;
                case "selectedItem":
                    this._selectionChangedAction({
                        selectedItem: args.value
                    });
                    break;
                default:
                    this.callBase(args)
            }
        }
    }).include(DataExpressionMixin);
    registerComponent("dxDropDownList", DropDownList);
    module.exports = DropDownList
});
